# Software License Agreement (BSD License)
#
# Copyright (c) 2009, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import sys
import time
import traceback
import unittest

import genpy
import genpy.message
from genpy import Duration
from genpy import Time

from nose.tools import nottest

import yaml

# Not much to test, just tripwires


class MessageTest(unittest.TestCase):

    def xtest_check_types_Header(self):
        # #2128: test that check_types works with a Header
        # #message. This is a weird case because Header has an aliased
        # #type and is the only ROS type for which that is true
        from test_roslib_comm.msg import HeaderTest
        x = HeaderTest()
        x._check_types()

    @nottest
    def test_Message_check_types(self):
        # this test is disabled since genpy does not depend on std_msgs and can therefore not use it in the test

        # test on a generated message
        # - use UInt16MultiArray because it has an embedded MultiArrayLayout
        from std_msgs.msg import String, UInt16MultiArray, MultiArrayLayout, MultiArrayDimension
        from genpy import SerializationError
        # not checking overflow in this test
        correct = [String(), String('foo'), String(''), String(data='data'),
                   UInt16MultiArray(),
                   UInt16MultiArray(MultiArrayLayout(), []),
                   UInt16MultiArray(MultiArrayLayout(data_offset=1), [1, 2, 3]),
                   UInt16MultiArray(layout=MultiArrayLayout(data_offset=1)),
                   UInt16MultiArray(layout=MultiArrayLayout(dim=[])),
                   UInt16MultiArray(layout=MultiArrayLayout(dim=[MultiArrayDimension()])),
                   UInt16MultiArray(data=[1, 2, 3]),
                   ]
        for t in correct:
            t._check_types()
        for t in correct:
            try:
                t._check_types(exc=Exception())
                self.fail('should have raised wrapped exc')
            except SerializationError:
                pass

        wrong = [String(1), String(data=1),
                 UInt16MultiArray(1, []),
                 UInt16MultiArray(MultiArrayLayout(), 1),
                 UInt16MultiArray(String(), []),
                 UInt16MultiArray(layout=MultiArrayLayout(dim=[1])),
                 UInt16MultiArray(layout=MultiArrayLayout(data_offset='')),
                 ]
        for t in wrong:
            try:
                t._check_types()
                self.fail('should have raised')
            except SerializationError:
                pass

    def test_fill_message_args_fixed_size_3785(self):
        # mainly regression tests for #3785: filling fixed-sized arrays

        from genpy.message import fill_message_args
        from genpy import Message, MessageException

        from genpy.msg import TestPrimitiveArray

        class M1(Message):
            __slots__ = ['int']
            _slot_types = ['int32']

            def __init__(self, int_=None):
                if int_ is None:
                    int_ = 0
                self.int = int_

        msg = M1()
        fill_message_args(msg, [{'int': 16}])
        assert msg == M1(16)
        fill_message_args(msg, [17])
        assert msg == M1(17)
        msg = M1()
        fill_message_args(msg, [{}])
        assert msg == M1()

        msg = TestPrimitiveArray()
        fill_message_args(msg, [{}])
        assert msg == TestPrimitiveArray()

        msg = TestPrimitiveArray()
        # - test fill with non-fixed size
        fill_message_args(msg, [{'ints': [1, 2, 3, 4], 'strings': ['a', 'b', 'c']}])
        assert msg == TestPrimitiveArray(ints=[1, 2, 3, 4], strings=['a', 'b', 'c'])
        # - test fill with fixed size
        msg = TestPrimitiveArray()
        fill_message_args(msg, [{'fixed_ints': [1, 2, 3, 4], 'fixed_strings': ['a', 'b', 'c', 'd']}])
        assert msg == TestPrimitiveArray(fixed_ints=[1, 2, 3, 4], fixed_strings=['a', 'b', 'c', 'd'])
        # - underflow
        try:
            fill_message_args(msg, [{'fixed_ints': [1, 2, 3], 'fixed_strings': ['a', 'b', 'c', 'd']}])
            assert False, 'should have failed'
        except MessageException:
            pass
        try:
            fill_message_args(msg, [{'fixed_ints': [1, 2, 3, 4], 'fixed_strings': ['a', 'b', 'c']}])
            assert False, 'should have failed'
        except MessageException:
            pass
        # - overflow
        try:
            fill_message_args(msg, [{'fixed_ints': [1, 2, 3, 4, 5], 'fixed_strings': ['a', 'b', 'c', 'd']}])
            assert False, 'should have failed'
        except MessageException:
            pass
        try:
            fill_message_args(msg, [{'fixed_ints': [1, 2, 3, 4], 'fixed_strings': ['a', 'b', 'c', 'd', 'e']}])
            assert False, 'should have failed'
        except MessageException:
            pass

        # test fixed-size message arrays
        from genpy.msg import TestMsgArray, TestString
        msg = TestMsgArray()
        fill_message_args(msg, [{}])
        assert msg == TestMsgArray()

        msg = TestMsgArray()
        assert len(msg.strings) == 0
        assert len(msg.fixed_strings) == 1
        # - test fill with non-fixed size
        fill_message_args(msg, [{'strings': [['a'], ['b']]}])
        assert msg == TestMsgArray(strings=[TestString('a'), TestString('b')])
        # - test fill with fixed size
        msg = TestMsgArray()
        fill_message_args(msg, [{'fixed_strings': [['a'], ]}])
        assert msg == TestMsgArray(fixed_strings=[TestString('a')])
        # overflow
        msg = TestMsgArray()
        try:
            fill_message_args(msg, [{'fixed_strings': [['a'], ['b']]}])
            assert False, 'should have failed'
        except MessageException:
            pass
        # underflow
        msg = TestMsgArray()
        try:
            fill_message_args(msg, [{'fixed_strings': []}])
            assert False, 'should have failed'
        except MessageException:
            pass

    def test_fill_message_args_embed_time(self):
        from genpy import Time
        from genpy.message import fill_message_args
        from genpy.msg import TestFillEmbedTime

        # test fill_message_args with embeds and time vals
        # time t
        # duration d
        # std_msgs/String str_msg
        # std_msgs/String[] str_msg_array
        # int32 i32

        m = TestFillEmbedTime()
        fill_message_args(m, [{}])
        self.assertEqual(m.t, Time())
        self.assertEqual(m.d, Duration())
        self.assertEqual(m.str_msg.data, '')
        self.assertEqual(m.str_msg_array, [])
        self.assertEqual(m.i32, 0)

        # list tests
        # - these should be equivalent
        equiv = [
            [[10, 20], [30, 40], ['foo'], [['bar'], ['baz']], 32],
            [{'secs': 10, 'nsecs': 20}, {'secs': 30, 'nsecs': 40}, ['foo'], [['bar'], ['baz']], 32],
            [[10, 20], [30, 40], {'data': 'foo'}, [['bar'], ['baz']], 32],
            [[10, 20], [30, 40], ['foo'], [{'data': 'bar'}, {'data': 'baz'}], 32],

            [{'t': [10, 20], 'd': [30, 40], 'str_msg': {'data': 'foo'}, 'str_msg_array': [{'data': 'bar'}, {'data': 'baz'}], 'i32': 32}],
            [{'t': {'secs': 10, 'nsecs': 20}, 'd': [30, 40], 'str_msg': {'data': 'foo'}, 'str_msg_array': [{'data': 'bar'}, {'data': 'baz'}], 'i32': 32}],
            ]
        for test in equiv:
            m = TestFillEmbedTime()
            try:
                fill_message_args(m, test)
            except Exception:
                self.fail('failed to fill with : %s\n%s' % (str(test), traceback.format_exc()))

            self.assertEqual(m.t, Time(10, 20))
            self.assertEqual(m.d, Duration(30, 40))
            self.assertEqual(m.str_msg.data, 'foo')
            self.assertEqual(len(m.str_msg_array), 2, m.str_msg_array)
            self.assertEqual(m.str_msg_array[0].data, 'bar')
            self.assertEqual(m.str_msg_array[1].data, 'baz')
            self.assertEqual(m.i32, 32)
        # test creation of Time/Duration from single number representation, which is necessary for

        # yaml single-number support
        # - cannot include in tests above as conversion from integer is lossy
        m = TestFillEmbedTime()
        fill_message_args(m, [10000000020, 30000000040, ['foo'], [['bar'], ['baz']], 32])
        self.assertEqual(10, m.t.secs)
        self.assertTrue(abs(20 - m.t.nsecs) < 2)
        self.assertEqual(30, m.d.secs)
        self.assertTrue(abs(40 - m.d.nsecs) < 2)
        self.assertEqual(len(m.str_msg_array), 2, m.str_msg_array)
        self.assertEqual(m.str_msg_array[0].data, 'bar')
        self.assertEqual(m.str_msg_array[1].data, 'baz')
        self.assertEqual(m.i32, 32)

        bad = [
            # underfill in sub-args
            [[10, 20], [30, 40], ['foo'], [['bar'], ['baz']]],
            [[10], [30, 40], ['foo'], [['bar'], ['baz']], 32],
            [[10, 20], [30], ['foo'], [['bar'], ['baz']], 32],
            [[10, 20], [30, 40], [], [['bar'], ['baz']], 32],
            [[10, 20], [30, 40], ['foo'], [['bar'], []], 32],

            # overfill
            [[10, 20], [30, 40], ['foo'], [['bar'], ['baz']], 32, 64],
            [[10, 20, 30], [30, 40], ['foo'], [['bar'], ['baz']], 32],
            [[10, 20], [30, 40, 50], ['foo'], [['bar'], ['baz']], 32],
            [[10, 20], [30, 40], ['foo', 'bar'], [['bar'], ['baz']], 32],
            [[10, 20], [30, 40], ['foo'], [['bar', 'baz'], ['baz']], 32],
            [[10, 20], [30, 40], ['foo'], [['bar'], ['baz', 'car']], 32],

            # invalid fields
            [{'secs': 10, 'nsecs': 20, 'foo': 1}, {'secs': 30, 'nsecs': 40}, ['foo'], [['bar'], ['baz']], 32],
            [{'secs': 10, 'nsecs': 20}, {'secs': 30, 'nsecs': 40, 'foo': 1}, ['foo'], [['bar'], ['baz']], 32],
            [[10, 20], [30, 40], {'data': 'foo', 'fata': 1}, [['bar'], ['baz']], 32],
            [[10, 20], [30, 40], ['foo'], [{'data': 'bar'}, {'beta': 'baz'}], 32],
            [{'t': [10, 20], 'd': [30, 40], 'str_msg': {'data': 'foo'}, 'str_msg_array': [{'data': 'bar'}, {'data': 'baz'}], 'i32': 32, 'i64': 64}],
            ]
        for b in bad:
            failed = True
            try:
                m = TestFillEmbedTime()
                fill_message_args(m, b)
            except genpy.MessageException:
                failed = False
            self.assertFalse(failed, 'fill_message_args should have failed: %s' % str(b))

    def test_fill_message_args_simple(self):
        from genpy.message import fill_message_args
        from genpy.msg import TestFillSimple
        # int32 i32
        # string str
        # int32[] i32_array
        # bool b

        for v in [[], {}]:
            try:
                fill_message_args(object(), v)
                self.fail('should have raised ValueError')
            except ValueError:
                pass
        try:
            m = TestFillSimple()
            # call underlying routine as the branch is not reachable from above
            genpy.message._fill_message_args(m, 1, {}, '')
            self.fail('should have raised ValueError for bad msg_args')
        except ValueError:
            pass

        simple_tests = [
            [1, 'foo', [], True],
            [1, 'foo', [1, 2, 3, 4], False],
        ]
        for test in simple_tests:
            m = TestFillSimple()
            fill_message_args(m, test)
            self.assertEqual(m.i32, test[0])
            self.assertEqual(m.str, test[1])
            self.assertEqual(m.i32_array, test[2])
            self.assertEqual(m.b, test[3])

        # test with dictionaries
        m = TestFillSimple()
        fill_message_args(m, [{}])
        self.assertEqual(m.i32, 0)
        self.assertEqual(m.str, '')
        self.assertEqual(m.i32_array, [])
        self.assertEqual(m.b, False)

        m = TestFillSimple()
        fill_message_args(m, [{'i32': 10}])
        self.assertEqual(m.i32, 10)
        self.assertEqual(m.str, '')
        self.assertEqual(m.i32_array, [])
        self.assertEqual(m.b, False)

        m = TestFillSimple()
        fill_message_args(m, [{'str': 'hello', 'i32_array': [1, 2, 3]}])
        self.assertEqual(m.i32, 0)
        self.assertEqual(m.str, 'hello')
        self.assertEqual(m.i32_array, [1, 2, 3])
        self.assertEqual(m.b, False)

        # fill_message_args currently does not type check
        bad = [
            # extra key
            [{'bad': 1, 'str': 'hello', 'i32_array': [1, 2, 3]}],
            # underfill
            [1, 'foo', [1, 2, 3]],
            # overfill
            [1, 'foo', [1, 2, 3], True, 1],
            # non-list value for list field
            [1, 'foo', 1, True],
            ]
        for b in bad:
            failed = True
            try:
                m = TestFillSimple()
                fill_message_args(m, b)
            except genpy.MessageException:
                failed = False
            self.assertFalse(failed, 'fill_message_args should have failed: %s' % str(b))

    def test_check_types_valid(self):
        """
        Test directly a bunch of valid combinations to check_types.

        check_type will throw an exception when it fails
        """
        import numpy as np
        genpy.message.check_type('test', 'uint8[]', 'byteDataIsAStringInPy')
        genpy.message.check_type('test', 'char[]', 'byteDataIsAStringInPy')
        genpy.message.check_type('test', 'uint8[]', [3, 4, 5])
        genpy.message.check_type('test', 'uint8[]', (3, 4, 5))
        genpy.message.check_type('test', 'char[]', [3, 4, 5])
        genpy.message.check_type('test', 'int32[]', [3, 4, 5])
        genpy.message.check_type('test', 'int32', -5)
        genpy.message.check_type('test', 'int64', -5)
        genpy.message.check_type('test', 'int16', -5)
        genpy.message.check_type('test', 'int8', -5)
        genpy.message.check_type('test', 'uint32', 5)
        genpy.message.check_type('test', 'uint64', 5)
        genpy.message.check_type('test', 'uint16', 5)
        genpy.message.check_type('test', 'uint8', 5)
        genpy.message.check_type('test', 'bool', True)
        genpy.message.check_type('test', 'bool', False)
        genpy.message.check_type('test', 'bool', 0)
        genpy.message.check_type('test', 'bool', 1)
        genpy.message.check_type('test', 'string', 'IAmAString')
        genpy.message.check_type('test', 'time', Time())
        genpy.message.check_type('test', 'duration', Duration(5))
        genpy.message.check_type('test', 'float32', 5)
        genpy.message.check_type('test', 'float32', 5.0)
        genpy.message.check_type('test', 'float32', np.int16(5))
        genpy.message.check_type('test', 'float32', np.float32(5.0))
        genpy.message.check_type('test', 'float32', float('inf'))
        genpy.message.check_type('test', 'float32', -float('inf'))
        genpy.message.check_type('test', 'float32', float('nan'))
        genpy.message.check_type('test', 'float32', -float('nan'))
        genpy.message.check_type('test', 'float32', 2147483647)   # resulting float: 2147483648.0
        genpy.message.check_type('test', 'float64', 5.0)
        genpy.message.check_type('test', 'float64', 1 + np.finfo(np.float64).eps)
        genpy.message.check_type('test', 'float64', np.float64(5.0))
        genpy.message.check_type('test', 'float64', np.int32(2147483647))
        # smallest representable float larger than 1.0 in builtin float type (64 bits). conversion to float32 loses precision.
        genpy.message.check_type('test', 'float32', float(1 + 2**-52))

    def test_check_types_invalid(self):
        from genpy import SerializationError
        self.assertRaises(SerializationError, genpy.message.check_type,
                          'test', 'int32[]', 'someString')
        self.assertRaises(SerializationError, genpy.message.check_type,
                          'test', 'uint32[]', [3, -2, 4])
        self.assertRaises(SerializationError, genpy.message.check_type,
                          'test', 'uint8', -2)
        self.assertRaises(SerializationError, genpy.message.check_type,
                          'test', 'uint16', -2)
        self.assertRaises(SerializationError, genpy.message.check_type,
                          'test', 'uint32', -2)
        self.assertRaises(SerializationError, genpy.message.check_type,
                          'test', 'uint64', -2)
        self.assertRaises(SerializationError, genpy.message.check_type,
                          'test', 'bool', -2)
        self.assertRaises(SerializationError, genpy.message.check_type,
                          'test', 'bool', 2)
        self.assertRaises(SerializationError, genpy.message.check_type,
                          'test', 'float64', 'someString')
        try:
            u = unichr(1234)
        except NameError:
            u = chr(1234)
        self.assertRaises(SerializationError, genpy.message.check_type,
                          'test', 'string', u)

    def test_Message(self):
        try:
            from cStringIO import StringIO
        except ImportError:
            from io import StringIO
        from genpy import Message, SerializationError
        self.assertTrue(isinstance(Message(), Message))
        m = Message()
        b = StringIO()
        m.serialize(b)
        m.deserialize('')

        # test args/keywords constructor
        try:
            Message(1, 2, 3, one=1, two=2, three=3)
            self.fail('Message should not allow *args and **kwds')
        except TypeError:
            pass
        try:
            Message()._get_types()
            self.fail('_get_types() should not be callable on abstract Message instance')
        except Exception:
            pass

        # test Empty message
        class M1(Message):
            __slots__ = []
            _slot_types = []

            def __init__(self, *args, **kwds):
                super(M1, self).__init__(*args, **kwds)

            def _get_types(self):
                return []

        # - test __str__ on empty
        self.assertEqual('', str(M1()))
        # - should not fail on default constructor
        M1()._check_types()
        # - must fail if provided an exception
        try:
            M1()._check_types(Exception('test'))
            self.fail('_check_types must fail if explicitly provided an exception')
        except SerializationError:
            pass

        # Test simple message with two fields
        class M2(Message):
            __slots__ = ['a', 'b']
            _slot_types = ['int32', 'int32']

            def _get_types(self): return ['int32', 'int32']

            def __init__(self, *args, **kwds):
                super(M2, self).__init__(*args, **kwds)
        self.assertEqual('a: 1\nb: 2', str(M2(1, 2)))
        # - test check types with two int type
        M2(1, 2)._check_types()
        M2(a=1, b=2)._check_types()
        invalid = [M2(a=1), M2('1', '2'), M2(1, '2'), M2(1., 2.), M2(None, 2)]
        for m in invalid:
            try:
                m._check_types()
                self.fail('check_types for %s should have failed' % m)
            except SerializationError:
                pass

        valid = [
            ((), {}, M1),
            ((), {}, M2),
            ((1, 2), {}, M2),
            ((), {'a': 1, 'b': 2}, M2),
            ((), {'a': 1}, M2), ((), {'b': 2}, M2),
            ]
        invalid = [
            ((1,), {}, M1),
            ((), {'one': 1}, M1),
            ((1), {}, M2), ((1, 2, 3), {}, M2),
            ((), {'c': 1}, M2), ((), {'a': 1, 'b': 2, 'c': 1}, M2),
            ]
        for args, kwds, cls in valid:
            cls(*args, **kwds)
        val = time.time()
        val2 = time.time()
        self.assertEqual(val, M2(val, 2).a)
        self.assertEqual(val, M2(1, val).b)
        self.assertEqual(val, M2(a=val).a)
        self.assertEqual(None, M2(a=val).b)
        self.assertEqual(None, M2(b=val).a)
        self.assertEqual(val, M2(b=val).b)
        self.assertEqual(val, M2(a=val, b=val2).a)
        self.assertEqual(val2, M2(a=val, b=val2).b)
        for args, kwds, cls in invalid:
            try:
                cls(*args, **kwds)
                self.fail('Message should have failed for cls[%s] *args[%s] and **kwds[%s]' % (cls, args, kwds))
            except Exception:
                pass

        # Test inequality of identical messages is False
        self.assertFalse(M2(a=1, b=2) != M2(a=1, b=2))

    def test_strify_message(self):
        # this is a bit overtuned, but it will catch regressions
        from genpy.message import Message, strify_message

        class M1(Message):
            __slots__ = []
            _slot_types = []

            def __init__(self):
                pass
        self.assertEqual('', strify_message(M1()))

        class M2(Message):
            __slots__ = ['str', 'int', 'float', 'bool', 'list']
            _slot_types = ['string', 'int32', 'float32', 'bool', 'int32[]']

            def __init__(self, str_, int_, float_, bool_, list_):
                self.str = str_
                self.int = int_
                self.float = float_
                self.bool = bool_
                self.list = list_

        self.assertEqual("""str: "string"
int: 123456789101112
float: 5678.0
bool: True
list: [1, 2, 3]""", strify_message(M2('string', 123456789101112, 5678., True, [1, 2, 3])))

        self.assertEqual("""str: ''
int: -1
float: 0.0
bool: False
list: []""", strify_message(M2('', -1, 0., False, [])))

        class M3(Message):
            __slots__ = ['m2']
            _slot_types = ['M1']

            def __init__(self, m2):
                self.m2 = m2
        self.assertEqual('m2: ' + """
  str: "string"
  int: -1
  float: 0.0
  bool: False
  list: []""", strify_message(M3(M2('string', -1, 0., False, []))))

        # test array of Messages field
        class M4(Message):
            __slots__ = ['m2s']
            _slot_types = ['M2[]']

            def __init__(self, m2s):
                self.m2s = m2s

        self.assertEqual('m2s: ' + """
  - """ + """
    str: "string"
    int: 1234
    float: 5678.0
    bool: True
    list: [1, 2, 3]
  - """ + """
    str: "string"
    int: -1
    float: 0.0
    bool: False
    list: []""", strify_message(M4([
                        M2('string', 1234, 5678., True, [1, 2, 3]),
                        M2('string', -1, 0., False, []),
                        ])))
        # test Time and Duration
        from genpy import Time, Duration

        class M5(Message):
            __slots__ = ['t', 'd']
            _slot_types = ['time', 'duration']

            def __init__(self, t, d):
                self.t = t
                self.d = d
        self.assertEqual('t: ' + """
  secs: 987
  nsecs:       654
d: """ + """
  secs: 123
  nsecs:       456""", strify_message(M5(Time(987, 654), Duration(123, 456))))

        # test final clause of strify -- str anything that isn't recognized
        if sys.hexversion > 0x03000000:  # Python3
            self.assertEqual('{1}', strify_message({1}))
        else:
            self.assertEqual('set([1])', strify_message({1}))

    def test_strify_yaml(self):
        def roundtrip(m):
            yaml_text = strify_message(m)
            print(yaml_text)
            loaded = yaml.safe_load(yaml_text)
            print('loaded', loaded)
            new_inst = m.__class__()
            if loaded is not None:
                fill_message_args(new_inst, [loaded])
            else:
                fill_message_args(new_inst, [])
            return new_inst

        # test YAML roundtrip. strify_message doesn't promise this
        # yet, but want to use it in this way in some demo toolchains
        from genpy.message import Message, strify_message, fill_message_args

        class M1(Message):
            __slots__ = []
            _slot_types = []

            def __init__(self):
                pass
        self.assertEqual(M1(), roundtrip(M1()))

        class M2(Message):
            __slots__ = ['str', 'int', 'float', 'bool', 'list', 'str_list']
            _slot_types = ['string', 'int32', 'float32', 'bool', 'int32[]', 'string[]']

            def __init__(self, str_=None, int_=None, float_=None, bool_=None, list_=None, str_list_=None):
                self.str = str_
                self.int = int_
                self.float = float_
                self.bool = bool_
                self.list = list_
                self.str_list = str_list_

        # test with empty string and empty list
        val = M2('foo\nbar', -1, 0., False, [], ['', ''])
        self.assertEqual(val, roundtrip(val))

        multiline_str = """
foo bar
foo # bar
foo ' bar
foo " bar
        """
        # test with strings and list of strings that need escaping
        val = M2(multiline_str, 123456789101112, 5678., True, [1, 2, 3], [multiline_str, 'foo \n # bar\n', '"foo \' bar"', '"foo" \" # \' \" bar'])
        self.assertEqual(val, roundtrip(val))

        class M3(Message):
            __slots__ = ['m2']
            _slot_types = ['test_roslib/M2']

            def __init__(self, m2=None):
                self.m2 = m2 or M2()

        # test with nested complex message
        val = M3(val)
        self.assertEqual(val, roundtrip(val))

    def test_check_type(self):
        # check_type() currently does not do harder checks like
        # type-checking class types.  as soon as it does, it will need
        # test to validate this.
        from genpy.message import check_type, SerializationError
        from genpy import Time, Duration
        valids = [
            ('byte', 1), ('byte', -1),
            ('string', ''), ('string', 'a string of text'),
            ('int32[]', []),
            ('int32[]', [1, 2, 3, 4]),
            ('time', Time()), ('time', Time.from_sec(1.0)),
            ('time', Time(10000)), ('time', Time(1000, -100)),
            ('duration', Duration()), ('duration', Duration()),
            ('duration', Duration(100)), ('duration', Duration(-100, -100)),
                  ]

        for t, v in valids:
            try:
                check_type('n', t, v)
            except Exception as e:
                traceback.print_exc()
                raise Exception('failure type[%s] value[%s]: %s' % (t, v, str(e)))

        invalids = [
            ('byte', 129), ('byte', -129), ('byte', 'byte'), ('byte', 1.0),
            ('string', 1),
            ('uint32', -1),
            ('int8', 112312), ('int8', -112312),
            ('uint8', -1), ('uint8', 112312),
            ('int32', '1'), ('int32', 1.),
            ('int32[]', 1), ('int32[]', [1., 2.]), ('int32[]', [1, 2.]),
            ('duration', 1), ('time', 1),
            ]
        for t, v in invalids:
            try:
                check_type('n', t, v)
                self.fail('check_type[%s, %s] should have failed' % (t, v))
            except SerializationError:
                pass

    def test_serialize_exception_msg(self):
        from genpy.message import fill_message_args
        from genpy.msg import TestStringFloat
        try:
            from cStringIO import StringIO
        except ImportError:
            from io import BytesIO as StringIO

        buff = StringIO()

        m = TestStringFloat()
        fill_message_args(m, ['foo', 1.0])
        self.assertEqual(m.data, 'foo')
        self.assertEqual(m.float, 1.0)

        # should serialize fine with float
        m.serialize(buff)

        # setting a string instead should fail with correct message
        m.float = '1.0'
        try:
            m.serialize(buff)
            assert False, 'This should have raised a genpy.SerializationError'
        except genpy.SerializationError as e:
            self.assertEqual(str(e), "field float must be float type")
        except Exception:
            assert False, 'This should have raised a genpy.SerializationError instead'

    @unittest.skipIf(sys.hexversion < 0x03000000, "Python 3 only test")
    def test_deserialize_unicode_error(self):
        from genpy.msg import TestString, TestMsgArray

        m = TestString()
        buff = b'\x00\x00\x00\x04\xF0\x9F\x92\xA9'
        self.assertEqual(m.deserialize(buff).data, b'\xF0\x9F\x92\xA9'.decode())

        m = TestString()
        buff = b'\x00\x00\x00\x04\x41\xff\xfe\x42'
        with self.assertLogs('rosout', level='ERROR') as cm:
            self.assertEqual(m.deserialize(buff).data, 'A\ufffd\ufffdB')
            self.assertEqual(m.deserialize(buff).data, 'A\ufffd\ufffdB')
            self.assertEqual(len(cm.output), 1)
            self.assertIn("Characters replaced when decoding message genpy/TestString (will print only once)", cm.output[0])

        m = TestMsgArray()
        buff = b'\x00\x00\x00\x00\x00\x00\x00\x04\x41\xff\xfe\x42'
        with self.assertLogs('rosout', level='ERROR') as cm:
            self.assertEqual(m.deserialize(buff).fixed_strings[0].data, 'A\ufffd\ufffdB')
            self.assertEqual(m.deserialize(buff).fixed_strings[0].data, 'A\ufffd\ufffdB')
            self.assertEqual(len(cm.output), 1)
            self.assertIn("Characters replaced when decoding message genpy/TestMsgArray (will print only once)", cm.output[0])



def test_reduce_pattern():
    from genpy.generate_struct import reduce_pattern
    tests = [
        ('', ''),
        ('hhhh', '4h'),
        ('hhhhi', '4hi'),
        ('hhhhiiiibbb', '4h4i3b'),
        ('1h2h3h', '1h2h3h'),
        ('hIi', 'hIi'),
        ('66h', '66h'),
        ('%ss', '%ss'),  # don't reduce strings with format chars in them
        ('<I', '<I'),
        ('<11s', '<11s'),
        ]
    for input_, result in tests:
        assert result == reduce_pattern(input_)


def test_pack():
    from genpy.generate_struct import pack
    assert 'buff.write(_get_struct_3lL3bB().pack(foo, bar))' == pack('lllLbbbB', 'foo, bar')


def test_pack2():
    from genpy.generate_struct import pack2
    assert 'buff.write(struct.Struct(patt_name).pack(foo, bar))' == pack2('patt_name', 'foo, bar')


def test_unpack():
    from genpy.generate_struct import unpack
    assert 'var_x = _get_struct_I3if2I().unpack(bname)' == unpack('var_x', 'IiiifII', 'bname')


def test_unpack2():
    from genpy.generate_struct import unpack2
    assert 'x = struct.unpack(patt, b)' == unpack2('x', 'patt', 'b')


def test_unpack3():
    from genpy.generate_struct import unpack3
    assert 'x = s.unpack(b)' == unpack3('x', 's', 'b')


def test_compute_struct_pattern():
    from genpy.generate_struct import compute_struct_pattern
    assert compute_struct_pattern(None) is None
    assert compute_struct_pattern([]) is None
    # string should immediately bork any simple types
    assert compute_struct_pattern(['string']) is None
    assert compute_struct_pattern(['uint32', 'string']) is None
    assert compute_struct_pattern(['string', 'int32']) is None
    # array types should not compute
    assert compute_struct_pattern(['uint32[]']) is None
    assert compute_struct_pattern(['uint32[1]']) is None

    assert 'B' == compute_struct_pattern(['uint8'])
    assert 'BB' == compute_struct_pattern(['uint8', 'uint8'])
    assert 'B' == compute_struct_pattern(['char'])
    assert 'BB' == compute_struct_pattern(['char', 'char'])
    assert 'b' == compute_struct_pattern(['byte'])
    assert 'bb' == compute_struct_pattern(['byte', 'byte'])
    assert 'b' == compute_struct_pattern(['int8'])
    assert 'bb' == compute_struct_pattern(['int8', 'int8'])
    assert 'H' == compute_struct_pattern(['uint16'])
    assert 'HH' == compute_struct_pattern(['uint16', 'uint16'])
    assert 'h' == compute_struct_pattern(['int16'])
    assert 'hh' == compute_struct_pattern(['int16', 'int16'])
    assert 'I' == compute_struct_pattern(['uint32'])
    assert 'II' == compute_struct_pattern(['uint32', 'uint32'])
    assert 'i' == compute_struct_pattern(['int32'])
    assert 'ii' == compute_struct_pattern(['int32', 'int32'])
    assert 'Q' == compute_struct_pattern(['uint64'])
    assert 'QQ' == compute_struct_pattern(['uint64', 'uint64'])
    assert 'q' == compute_struct_pattern(['int64'])
    assert 'qq' == compute_struct_pattern(['int64', 'int64'])
    assert 'f' == compute_struct_pattern(['float32'])
    assert 'ff' == compute_struct_pattern(['float32', 'float32'])
    assert 'd' == compute_struct_pattern(['float64'])
    assert 'dd' == compute_struct_pattern(['float64', 'float64'])

    assert 'bBhHiIqQfd' == compute_struct_pattern(['int8', 'uint8', 'int16', 'uint16', 'int32', 'uint32', 'int64', 'uint64', 'float32', 'float64'])



import os
import sys

pkg_path = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
sys.path.insert(0, os.path.join(pkg_path, 'src'))

from genpy.generator import MsgGenerator  # noqa
from genpy.genpy_main import genmain  # noqa


def generate_test_messages(msg_files):
    test_msg_path = os.path.join(pkg_path, 'test/msg')
    msg_files = []
    for filename in os.listdir(test_msg_path):
        if filename.endswith('.msg'):
            msg_files.append(filename)

    src_msg_path = os.path.join(pkg_path, 'src/genpy/msg')
    if not os.path.exists(src_msg_path):
        os.makedirs(src_msg_path)

    with open(os.path.join(src_msg_path, '__init__.py'), 'w') as h:
        for msg_file in msg_files:
            h.write('from ._%s import *\n' % msg_file[:-4])

    for msg_file in msg_files:
        argv = [
            'genmsg_py.py',
            '-p', 'genpy',
            '-Igenpy:%s/test/msg' % pkg_path,
            '-o', '%s/src/genpy/msg' % pkg_path,
            '%s/test/msg/%s' % (pkg_path, msg_file),
        ]
        try:
            genmain(argv, 'genmsg_py.py', MsgGenerator())
        except SystemExit as e:
            if e.code:
                raise


if __name__ == '__main__':
    generate_test_messages(sys.argv[1:])



# Software License Agreement (BSD License)
#
# Copyright (c) 2008, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

try:
    from cStringIO import StringIO
except ImportError:
    from io import BytesIO as StringIO
import sys
import time


def test_generate_dynamic():
    import genpy
    from genpy.dynamic import generate_dynamic
    msgs = generate_dynamic('gd_msgs/EasyString', 'string data\n')
    assert ['gd_msgs/EasyString'] == list(msgs.keys())
    m_cls = msgs['gd_msgs/EasyString']
    m_instance = m_cls()
    m_instance.data = 'foo'
    buff = StringIO()
    m_instance.serialize(buff)
    m_instance2 = m_cls().deserialize(buff.getvalue())
    assert m_instance == m_instance2

    try:
        msgs = generate_dynamic('gd_msgs/MyAcceleration', 'float32 acceleration  # in m/s\xc2\xb2\n')
    except UnicodeDecodeError:
        assert False, "Can't handle UTF-8 in comments"

    try:
        char = unichr
    except NameError:
        char = chr
    m_instance.data = 'foo' + char(1234)
    buff = StringIO()
    m_instance.serialize(buff)
    m_instance2 = m_cls().deserialize(buff.getvalue())
    if sys.hexversion < 0x03000000:
        # python 2 requires manual decode into unicode
        m_instance2.data = m_instance2.data.decode('utf-8')
    assert m_instance == m_instance2

    # 'probot_msgs' is a test for #1183, failure if the package no longer exists
    msgs = generate_dynamic('gd_msgs/MoveArmState', """Header header
probot_msgs/ControllerStatus status

#Current arm configuration
probot_msgs/JointState[] configuration
#Goal arm configuration
probot_msgs/JointState[] goal

================================================================================
MSG: std_msgs/Header
#Standard metadata for higher-level flow data types
#sequence ID: consecutively increasing ID
uint32 seq
#Two-integer timestamp that is expressed as:
# * stamp.secs: seconds (stamp_secs) since epoch
# * stamp.nsecs: nanoseconds since stamp_secs
# time-handling sugar is provided by the client library
time stamp
#Frame this data is associated with
# 0: no frame
# 1: global frame
string frame_id

================================================================================
MSG: probot_msgs/ControllerStatus
# This message defines the expected format for Controller Status messages
# Embed this in the feedback state message of highlevel controllers
byte UNDEFINED=0
byte SUCCESS=1
byte ABORTED=2
byte PREEMPTED=3
byte ACTIVE=4

# Status of the controller = {UNDEFINED, SUCCESS, ABORTED, PREEMPTED, ACTIVE}
byte value

#Comment for debug
string comment
================================================================================
MSG: probot_msgs/JointState
string name
float64 position
float64 velocity
float64 applied_effort
float64 commanded_effort
byte is_calibrated

""")
    assert {'gd_msgs/MoveArmState', 'probot_msgs/JointState', 'probot_msgs/ControllerStatus', 'std_msgs/Header'} == set(msgs.keys())
    m_instance1 = msgs['std_msgs/Header']()  # make sure default constructor works
    m_instance2 = msgs['std_msgs/Header'](stamp=genpy.Time.from_sec(time.time()), frame_id='foo-%s' % time.time(), seq=12390)
    _test_ser_deser(m_instance2, m_instance1)

    m_instance1 = msgs['probot_msgs/ControllerStatus']()
    m_instance2 = msgs['probot_msgs/ControllerStatus'](value=4, comment=str(time.time()))
    d = {'UNDEFINED': 0, 'SUCCESS': 1, 'ABORTED': 2, 'PREEMPTED': 3, 'ACTIVE': 4}
    for k, v in d.items():
        assert v == getattr(m_instance1, k)
    _test_ser_deser(m_instance2, m_instance1)

    m_instance1 = msgs['probot_msgs/JointState']()
    m_instance2 = msgs['probot_msgs/JointState'](position=time.time(), velocity=time.time(), applied_effort=time.time(), commanded_effort=time.time(), is_calibrated=2)
    _test_ser_deser(m_instance2, m_instance1)

    m_instance1 = msgs['gd_msgs/MoveArmState']()
    js = msgs['probot_msgs/JointState']
    config = []
    goal = []
    # generate some data for config/goal
    for i in range(0, 10):
        config.append(js(position=time.time(), velocity=time.time(), applied_effort=time.time(), commanded_effort=time.time(), is_calibrated=2))
        goal.append(js(position=time.time(), velocity=time.time(), applied_effort=time.time(), commanded_effort=time.time(), is_calibrated=2))
    m_instance2 = msgs['gd_msgs/MoveArmState'](header=msgs['std_msgs/Header'](),
                                               status=msgs['probot_msgs/ControllerStatus'](),
                                               configuration=config, goal=goal)
    _test_ser_deser(m_instance2, m_instance1)


def _test_ser_deser(m_instance1, m_instance2):
    buff = StringIO()
    m_instance1.serialize(buff)
    m_instance2.deserialize(buff.getvalue())
    assert m_instance1 == m_instance2


def test_serialize_exception():
    import genpy
    from genpy.dynamic import generate_dynamic
    msgs = generate_dynamic('gd_msgs/EasyInt32', 'int32 data\n')
    assert ['gd_msgs/EasyInt32'] == list(msgs.keys())
    m_cls = msgs['gd_msgs/EasyInt32']
    m_instance = m_cls()
    m_instance.data = '1'  # the type is incorrect
    buff = StringIO()
    try:
        m_instance.serialize(buff)
        assert False, 'This should have raised a genpy.SerializationError'
    except genpy.SerializationError:
        pass
    except Exception:
        assert False, 'This should have raised a genpy.SerializationError instead'



import traceback
import unittest


class MessageTest(unittest.TestCase):

    def test_python_safe_message(self):
        # simple test for make_python_safe
        try:
            # this import will fail, if the data field "from" is not properly properly renames, since "from" is a reserved keyword in Python
            from genpy.msg import TestPythonSafe  # noqa: F401
        except Exception:
            # assert False, "should have raised"
            self.fail("failed to import message type 'TestPythonSafe':\n%s" % (traceback.format_exc()))

    def test_python_safe_message_with_subfields(self):
        # regression test for https://github.com/ros/genpy/issues/68
        try:
            # this import will fail, if the make_python_safe function is not properly used on all subfields of the message
            from genpy.msg import TestPythonSafeSubfields  # noqa: F401
        except Exception:
            self.fail("failed to import message type 'TestPythonSafeSubfields':\n%s" % (traceback.format_exc()))



# Software License Agreement (BSD License)
#
# Copyright (c) 2008, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os

import genmsg.msgs
from genmsg.msg_loader import MsgContext
from genmsg.msgs import MsgSpec


def get_test_dir():
    return os.path.abspath(os.path.join(os.path.dirname(__file__), 'files'))


def test_is_special():
    import genpy.generator
    for t in ['time', 'duration', 'Header']:
        assert genpy.generator.is_special(t)


def test_Simple():
    import genpy.generator
    val = genpy.generator.get_special('time').import_str
    assert 'import genpy' == val, val
    assert 'import genpy' == genpy.generator.get_special('duration').import_str
    assert 'import std_msgs.msg' == genpy.generator.get_special('Header').import_str

    assert 'genpy.Time()' == genpy.generator.get_special('time').constructor
    assert 'genpy.Duration()' == genpy.generator.get_special('duration').constructor
    assert 'std_msgs.msg._Header.Header()' == genpy.generator.get_special('Header').constructor

    assert 'self.foo.canon()' == genpy.generator.get_special('time').get_post_deserialize('self.foo')
    assert 'bar.canon()' == genpy.generator.get_special('time').get_post_deserialize('bar')
    assert 'self.foo.canon()' == genpy.generator.get_special('duration').get_post_deserialize('self.foo')
    assert genpy.generator.get_special('Header').get_post_deserialize('self.foo') is None


def test_compute_post_deserialize():
    import genpy.generator
    assert 'self.bar.canon()' == genpy.generator.compute_post_deserialize('time', 'self.bar')
    assert 'self.bar.canon()' == genpy.generator.compute_post_deserialize('duration', 'self.bar')
    assert genpy.generator.compute_post_deserialize('Header', 'self.bar') is None

    assert genpy.generator.compute_post_deserialize('int8', 'self.bar') is None
    assert genpy.generator.compute_post_deserialize('string', 'self.bar') is None


def test_flatten():
    from genpy.generator import flatten
    msg_context = MsgContext.create_default()

    simple = MsgSpec(['string'], ['data'], [], 'string data\n', 'simple/String')
    simple2 = MsgSpec(['string', 'int32'], ['data', 'data2'], [], 'string data\nint32 data2\n', 'simpe/Data2')
    assert simple == flatten(msg_context, simple)
    assert simple2 == flatten(msg_context, simple2)

    b1 = MsgSpec(['int8'], ['data'], [], 'X', 'f_msgs/Base')
    b2 = MsgSpec(['f_msgs/Base'], ['data'], [], 'X', 'f_msgs/Base2')
    b3 = MsgSpec(['f_msgs/Base2', 'f_msgs/Base2'], ['data3', 'data4'], [], 'X', 'f_msgs/Base3')
    b4 = MsgSpec(['f_msgs/Base3', 'f_msgs/Base3'], ['dataA', 'dataB'], [], 'X', 'f_msgs/Base4')

    msg_context.register('f_msgs/Base', b1)
    msg_context.register('f_msgs/Base2', b2)
    msg_context.register('f_msgs/Base3', b3)
    msg_context.register('f_msgs/Base4', b4)

    assert MsgSpec(['int8'], ['data.data'], [], 'X', 'f_msgs/Base2') == flatten(msg_context, b2)
    assert MsgSpec(['int8', 'int8'], ['data3.data.data', 'data4.data.data'], [], 'X', 'f_msgs/Base3') == flatten(msg_context, b3)
    assert MsgSpec(
        ['int8', 'int8', 'int8', 'int8'],
        ['dataA.data3.data.data', 'dataA.data4.data.data', 'dataB.data3.data.data', 'dataB.data4.data.data'],
        [], 'X', 'f_msgs/Base4') == flatten(msg_context, b4)


def test_flatten_array_objects():
    # make sure array of types don't flatten
    from genpy.generator import flatten
    msg_context = MsgContext.create_default()

    b1 = MsgSpec(['int8'], ['data'], [], 'X', 'f_msgs/Base')
    b5 = MsgSpec(['f_msgs/Base[]'], ['data'], [], 'X', 'f_msgs/Base5')

    msg_context.register('f_msgs/Base', b1)
    msg_context.register('f_msgs/Base5', b5)
    assert b5 == flatten(msg_context, b5)


def test_default_value():
    from genpy.generator import default_value
    msg_context = MsgContext.create_default()

    msg_context.register('fake_msgs/String', MsgSpec(['string'], ['data'], [], 'string data\n', 'fake_msgs/String'))
    msg_context.register('fake_msgs/ThreeNums', MsgSpec(['int32', 'int32', 'int32'], ['x', 'y', 'z'], [], 'int32 x\nint32 y\nint32 z\n', 'fake_msgs/ThreeNums'))

    # trip-wire: make sure all builtins have a default value
    for t in genmsg.msgs.BUILTIN_TYPES:
        assert type(default_value(msg_context, t, 'roslib')) == str

    # simple types first
    for t in ['uint8', 'int8', 'uint16', 'int16', 'uint32', 'int32', 'uint64', 'int64', 'byte', 'char']:
        assert '0' == default_value(msg_context, t, 'std_msgs')
        assert '0' == default_value(msg_context, t, 'roslib')
    for t in ['float32', 'float64']:
        assert '0.' == default_value(msg_context, t, 'std_msgs')
        assert '0.' == default_value(msg_context, t, 'roslib')
    assert "''" == default_value(msg_context, 'string', 'roslib')

    # builtin specials
    assert 'genpy.Time()' == default_value(msg_context, 'time', 'roslib')
    assert 'genpy.Duration()' == default_value(msg_context, 'duration', 'roslib')
    assert 'std_msgs.msg._Header.Header()' == default_value(msg_context, 'Header', 'roslib')

    assert 'genpy.Time()' == default_value(msg_context, 'time', 'std_msgs')
    assert 'genpy.Duration()' == default_value(msg_context, 'duration', 'std_msgs')
    assert 'std_msgs.msg._Header.Header()' == default_value(msg_context, 'Header', 'std_msgs')

    # generic instances
    # - unregistered type
    assert default_value(msg_context, 'unknown_msgs/Foo', 'unknown_msgs') is None
    # - wrong context
    assert default_value(msg_context, 'ThreeNums', 'std_msgs') is None

    # - registered types
    assert 'fake_msgs.msg.String()' == default_value(msg_context, 'fake_msgs/String', 'std_msgs')
    assert 'fake_msgs.msg.String()' == default_value(msg_context, 'fake_msgs/String', 'fake_msgs')
    assert 'fake_msgs.msg.String()' == default_value(msg_context, 'String', 'fake_msgs')
    assert 'fake_msgs.msg.ThreeNums()' == default_value(msg_context, 'fake_msgs/ThreeNums', 'roslib')
    assert 'fake_msgs.msg.ThreeNums()' == default_value(msg_context, 'fake_msgs/ThreeNums', 'fake_msgs')
    assert 'fake_msgs.msg.ThreeNums()' == default_value(msg_context, 'ThreeNums', 'fake_msgs')

    # var-length arrays always default to empty arrays... except for byte and uint8 which are strings
    for t in ['int8', 'uint16', 'int16', 'uint32', 'int32', 'uint64', 'int64', 'float32', 'float64']:
        val = default_value(msg_context, t+'[]', 'std_msgs')
        assert '[]' == val, '[%s]: %s' % (t, val)
        assert '[]' == default_value(msg_context, t+'[]', 'roslib')

    assert b'' == eval(default_value(msg_context, 'uint8[]', 'roslib'))

    # fixed-length arrays should be zero-filled... except for byte and uint8 which are strings
    for t in ['float32', 'float64']:
        assert '[0.] * 3' == default_value(msg_context, t+'[3]', 'std_msgs')
        assert '[0.] * 1' == default_value(msg_context, t+'[1]', 'std_msgs')
    for t in ['int8', 'uint16', 'int16', 'uint32', 'int32', 'uint64', 'int64']:
        assert '[0] * 4' == default_value(msg_context, t+'[4]', 'std_msgs')
        assert '[0] * 1' == default_value(msg_context, t+'[1]', 'roslib')

    assert b'\0' == eval(default_value(msg_context, 'uint8[1]', 'roslib'))
    assert b'\0\0\0\0' == eval(default_value(msg_context, 'uint8[4]', 'roslib'))

    assert '[]' == default_value(msg_context, 'fake_msgs/String[]', 'std_msgs')
    assert '[fake_msgs.msg.String() for _ in range(2)]' == default_value(msg_context, 'fake_msgs/String[2]', 'std_msgs')


def test_make_python_safe():
    from genpy.generator import make_python_safe
    from genmsg.msgs import Constant
    s = MsgSpec(['int32', 'int32', 'int32', 'int32', 'int32', 'int32'], ['ok', 'if', 'self', 'fine', 'self.x', 'self.while'],
                [Constant('int32', 'if', '1', '1'), Constant('int32', 'okgo', '1', '1')],
                'x', 'test_msgs/Foo')
    s2 = make_python_safe(s)
    assert s != s2
    assert ['ok', 'if_', 'self_', 'fine', 'self.x', 'self.while_'] == s2.names, s2.names
    assert s2.types == s.types
    assert [Constant('int32', 'if_', '1', '1') == Constant('int32', 'okgo', '1', '1')], s2.constants
    assert s2.text == s.text


def test_compute_pkg_type():
    from genpy.generator import compute_pkg_type, MsgGenerationException
    try:
        compute_pkg_type('std_msgs', 'really/bad/std_msgs/String')
    except MsgGenerationException:
        pass
    assert ('std_msgs', 'String') == compute_pkg_type('std_msgs', 'std_msgs/String')
    assert ('std_msgs', 'String') == compute_pkg_type('foo', 'std_msgs/String')
    assert ('std_msgs', 'String') == compute_pkg_type('std_msgs', 'String')


def test_compute_import():
    import genpy.generator
    msg_context = MsgContext.create_default()

    assert [] == genpy.generator.compute_import(msg_context, 'foo', 'bar')
    assert [] == genpy.generator.compute_import(msg_context, 'foo', 'int32')

    msg_context.register('ci_msgs/Base', MsgSpec(['int8'], ['data'], [], 'int8 data\n', 'ci_msgs/Base'))
    msg_context.register('ci2_msgs/Base2', MsgSpec(['ci_msgs/Base'], ['data2'], [], 'ci_msgs/Base data2\n', 'ci2_msgs/Base2'))
    msg_context.register('ci3_msgs/Base3', MsgSpec(['ci2_msgs/Base2'], ['data3'], [], 'ci2_msgs/Base2 data3\n', 'ci3_msgs/Base3'))
    msg_context.register('ci4_msgs/Base', MsgSpec(['int8'], ['data'], [], 'int8 data\n', 'ci4_msgs/Base'))
    msg_context.register('ci4_msgs/Base4', MsgSpec(
        ['ci2_msgs/Base2', 'ci3_msgs/Base3'],
        ['data4a', 'data4b'],
        [], 'ci2_msgs/Base2 data4a\nci3_msgs/Base3 data4b\n', 'ci4_msgs/Base4'))

    msg_context.register('ci5_msgs/Base', MsgSpec(['time'], ['data'], [], 'time data\n', 'ci5_msgs/Base'))

    assert ['import ci_msgs.msg'] == genpy.generator.compute_import(msg_context, 'foo', 'ci_msgs/Base')
    assert ['import ci_msgs.msg'] == genpy.generator.compute_import(msg_context, 'ci_msgs', 'ci_msgs/Base')
    assert ['import ci2_msgs.msg', 'import ci_msgs.msg'] == genpy.generator.compute_import(msg_context, 'ci2_msgs', 'ci2_msgs/Base2')
    assert ['import ci2_msgs.msg', 'import ci_msgs.msg'] == genpy.generator.compute_import(msg_context, 'foo', 'ci2_msgs/Base2')
    assert ['import ci3_msgs.msg', 'import ci2_msgs.msg', 'import ci_msgs.msg'] == genpy.generator.compute_import(msg_context, 'ci3_msgs', 'ci3_msgs/Base3')

    assert {'import ci4_msgs.msg', 'import ci3_msgs.msg', 'import ci2_msgs.msg', 'import ci_msgs.msg'} == set(genpy.generator.compute_import(msg_context, 'foo', 'ci4_msgs/Base4'))
    assert {'import ci4_msgs.msg', 'import ci3_msgs.msg', 'import ci2_msgs.msg', 'import ci_msgs.msg'} == set(genpy.generator.compute_import(msg_context, 'ci4_msgs', 'ci4_msgs/Base4'))

    assert ['import ci4_msgs.msg'] == genpy.generator.compute_import(msg_context, 'foo', 'ci4_msgs/Base')
    assert ['import ci4_msgs.msg'] == genpy.generator.compute_import(msg_context, 'ci4_msgs', 'ci4_msgs/Base')
    assert ['import ci4_msgs.msg'] == genpy.generator.compute_import(msg_context, 'ci4_msgs', 'Base')

    assert ['import ci5_msgs.msg', 'import genpy'] == genpy.generator.compute_import(msg_context, 'foo', 'ci5_msgs/Base')


def test_get_registered_ex():
    import genpy.generator
    msg_context = MsgContext.create_default()
    s = MsgSpec(['string'], ['data'], [], 'string data\n', 'tgr_msgs/String')
    msg_context.register('tgr_msgs/String', s)
    assert s == genpy.generator.get_registered_ex(msg_context, 'tgr_msgs/String')
    try:
        genpy.generator.get_registered_ex(msg_context, 'bad_msgs/String')
    except genpy.generator.MsgGenerationException:
        pass


def test_compute_constructor():
    from genpy.generator import compute_constructor
    msg_context = MsgContext.create_default()
    msg_context.register('fake_msgs/String', MsgSpec(['string'], ['data'], [], 'string data\n', 'fake_msgs/String'))
    msg_context.register('fake_msgs/ThreeNums', MsgSpec(['int32', 'int32', 'int32'], ['x', 'y', 'z'], [], 'int32 x\nint32 y\nint32 z\n', 'fake_msgs/ThreeNums'))

    # builtin specials
    assert 'genpy.Time()' == compute_constructor(msg_context, 'roslib', 'time')
    assert 'genpy.Duration()' == compute_constructor(msg_context, 'roslib', 'duration')
    assert 'std_msgs.msg._Header.Header()' == compute_constructor(msg_context, 'std_msgs', 'Header')

    assert 'genpy.Time()' == compute_constructor(msg_context, 'std_msgs', 'time')
    assert 'genpy.Duration()' == compute_constructor(msg_context, 'std_msgs', 'duration')

    # generic instances
    # - unregistered type
    assert compute_constructor(msg_context, 'unknown_msgs', 'unknown_msgs/Foo') is None
    assert compute_constructor(msg_context, 'unknown_msgs', 'Foo') is None
    # - wrong context
    assert compute_constructor(msg_context, 'std_msgs', 'ThreeNums') is None

    # - registered types
    assert 'fake_msgs.msg.String()' == compute_constructor(msg_context, 'std_msgs', 'fake_msgs/String')
    assert 'fake_msgs.msg.String()' == compute_constructor(msg_context, 'fake_msgs', 'fake_msgs/String')
    assert 'fake_msgs.msg.String()' == compute_constructor(msg_context, 'fake_msgs', 'String')
    assert 'fake_msgs.msg.ThreeNums()' == compute_constructor(msg_context, 'fake_msgs', 'fake_msgs/ThreeNums')
    assert 'fake_msgs.msg.ThreeNums()' == compute_constructor(msg_context, 'fake_msgs', 'fake_msgs/ThreeNums')
    assert 'fake_msgs.msg.ThreeNums()' == compute_constructor(msg_context, 'fake_msgs', 'ThreeNums')


def test_len_serializer_generator():
    import genpy.generator
    # generator tests are mainly tripwires/coverage tests
    # Test Serializers
    # string serializer simply initializes local var
    g = genpy.generator.len_serializer_generator('foo', True, True)
    assert 'length = len(foo)' == '\n'.join(g)
    # array len serializer writes var
    g = genpy.generator.len_serializer_generator('foo', False, True)
    assert 'length = len(foo)\nbuff.write(_struct_I.pack(length))' == '\n'.join(g)

    # Test Deserializers
    val = """start = end
end += 4
(length,) = _struct_I.unpack(str[start:end])"""
    # string serializer and array serializer are identical
    g = genpy.generator.len_serializer_generator('foo', True, False)
    assert val == '\n'.join(g)
    g = genpy.generator.len_serializer_generator('foo', False, False)
    assert val == '\n'.join(g)


def test_string_serializer_generator():
    import genpy.generator
    # generator tests are mainly tripwires/coverage tests
    # Test Serializers
    g = genpy.generator.string_serializer_generator('foo', 'string', 'var_name', True)
    val = '\n'.join(g)
    assert """length = len(var_name)
if python3 or type(var_name) == unicode:
  var_name = var_name.encode('utf-8')
  length = len(var_name)
buff.write(struct.Struct('<I%ss'%length).pack(length, var_name))""" == val, val

    for t in ['uint8[]', 'byte[]', 'uint8[10]', 'byte[20]']:
        g = genpy.generator.string_serializer_generator('foo', 'uint8[]', 'b_name', True)
        assert """length = len(b_name)
# - if encoded as a list instead, serialize as bytes instead of string
if type(b_name) in [list, tuple]:
  buff.write(struct.Struct('<I%sB'%length).pack(length, *b_name))
else:
  buff.write(struct.Struct('<I%ss'%length).pack(length, b_name))""" == '\n'.join(g)

    # Test Deserializers
    val = """start = end
end += 4
(length,) = _struct_I.unpack(str[start:end])
start = end
end += length
if python3:
  var_name = str[start:end].decode('utf-8', 'rosmsg')
else:
  var_name = str[start:end]"""
    # string serializer and array serializer are identical
    g = genpy.generator.string_serializer_generator('foo', 'string', 'var_name', False)
    assert val == '\n'.join(g)


def test_array_serializer_generator_numpy():
    is_numpy = True
    from genpy.generator import array_serializer_generator
    d = os.path.join(get_test_dir(), 'array')
    # generator tests are mainly tripwires/coverage tests

    # array_serializer_generator(msg_context, package, type_, name, serialize, is_numpy):
    msg_context = MsgContext.create_default()

    # permutations: var length, unint8
    serialize = True
    result = array_serializer_generator(msg_context, '', 'uint8[]', 'data', serialize, is_numpy)
    compare_file(d, 'uint8_varlen_ser_np.txt', result)
    result = array_serializer_generator(msg_context, '', 'int16[]', 'data', serialize, is_numpy)
    compare_file(d, 'int16_varlen_ser_np.txt', result)
    result = array_serializer_generator(msg_context, '', 'uint8[8]', 'data', serialize, is_numpy)
    compare_file(d, 'uint8_fixed_ser_np.txt', result)
    result = array_serializer_generator(msg_context, '', 'int16[10]', 'data', serialize, is_numpy)
    compare_file(d, 'int16_fixed_ser_np.txt', result)

    serialize = False
    result = array_serializer_generator(msg_context, '', 'uint8[]', 'data', serialize, is_numpy)
    compare_file(d, 'uint8_varlen_deser_np.txt', result)
    result = array_serializer_generator(msg_context, '', 'int16[]', 'data', serialize, is_numpy)
    compare_file(d, 'int16_varlen_deser_np.txt', result)
    result = array_serializer_generator(msg_context, '', 'uint8[8]', 'data', serialize, is_numpy)
    compare_file(d, 'uint8_fixed_deser_np.txt', result)
    result = array_serializer_generator(msg_context, '', 'int16[10]', 'data', serialize, is_numpy)
    compare_file(d, 'int16_fixed_deser_np.txt', result)


def compare_file(d, filename, result):
    result = '\n'.join(result)
    expected = open(os.path.join(d, filename)).read().strip()
    assert result == expected, '\n[%s]\n[%s]' % (result, expected)


def exhaust(gen):
    list(gen)


def test_array_serializer_generator():
    from genmsg.msg_loader import load_msg_by_type
    from genpy.generator import array_serializer_generator, MsgGenerationException, reset_var
    d = os.path.join(get_test_dir(), 'array')
    # generator tests are mainly tripwires/coverage tests

    # array_serializer_generator(msg_context, package, type_, name, serialize, is_numpy):
    msg_context = MsgContext.create_default()
    # load in some objects
    search_path = {'foo': [d]}
    load_msg_by_type(msg_context, 'foo/Object', search_path)
    load_msg_by_type(msg_context, 'foo/ObjectArray', search_path)

    # permutations: var length, unint8
    is_numpy = False

    serialize = True
    result = array_serializer_generator(msg_context, '', 'uint8[]', 'data', serialize, is_numpy)
    compare_file(d, 'uint8_varlen_ser.txt', result)
    result = array_serializer_generator(msg_context, '', 'int16[]', 'data', serialize, is_numpy)
    compare_file(d, 'int16_varlen_ser.txt', result)
    result = array_serializer_generator(msg_context, '', 'uint8[8]', 'data', serialize, is_numpy)
    compare_file(d, 'uint8_fixed_ser.txt', result)
    result = array_serializer_generator(msg_context, '', 'int16[10]', 'data', serialize, is_numpy)
    compare_file(d, 'int16_fixed_ser.txt', result)
    result = array_serializer_generator(msg_context, '', 'bool[]', 'data', serialize, is_numpy)
    compare_file(d, 'bool_varlen_ser.txt', result)
    result = array_serializer_generator(msg_context, '', 'bool[3]', 'data', serialize, is_numpy)
    compare_file(d, 'bool_fixed_ser.txt', result)

    # for 'complex' types have to reset the variable generator
    reset_var()
    result = array_serializer_generator(msg_context, '', 'string[]', 'data', serialize, is_numpy)
    compare_file(d, 'string_varlen_ser.txt', result)
    reset_var()
    result = array_serializer_generator(msg_context, '', 'string[2]', 'data', serialize, is_numpy)
    compare_file(d, 'string_fixed_ser.txt', result)

    reset_var()
    result = array_serializer_generator(msg_context, 'foo', 'foo/Object[]', 'data', serialize, is_numpy)
    compare_file(d, 'object_varlen_ser.txt', result)
    reset_var()
    result = array_serializer_generator(msg_context, 'foo', 'foo/Object[3]', 'data', serialize, is_numpy)
    compare_file(d, 'object_fixed_ser.txt', result)

    serialize = False
    result = array_serializer_generator(msg_context, '', 'uint8[]', 'data', serialize, is_numpy)
    compare_file(d, 'uint8_varlen_deser.txt', result)
    result = array_serializer_generator(msg_context, '', 'int16[]', 'data', serialize, is_numpy)
    compare_file(d, 'int16_varlen_deser.txt', result)
    result = array_serializer_generator(msg_context, '', 'uint8[8]', 'data', serialize, is_numpy)
    compare_file(d, 'uint8_fixed_deser.txt', result)
    result = array_serializer_generator(msg_context, '', 'int16[10]', 'data', serialize, is_numpy)
    compare_file(d, 'int16_fixed_deser.txt', result)
    result = array_serializer_generator(msg_context, '', 'bool[]', 'data', serialize, is_numpy)
    compare_file(d, 'bool_varlen_deser.txt', result)
    result = array_serializer_generator(msg_context, '', 'bool[3]', 'data', serialize, is_numpy)
    compare_file(d, 'bool_fixed_deser.txt', result)

    # for 'complex' types have to reset the variable generator
    reset_var()
    result = array_serializer_generator(msg_context, '', 'string[]', 'data', serialize, is_numpy)
    compare_file(d, 'string_varlen_deser.txt', result)
    reset_var()
    result = array_serializer_generator(msg_context, '', 'string[2]', 'data', serialize, is_numpy)
    compare_file(d, 'string_fixed_deser.txt', result)

    reset_var()
    result = array_serializer_generator(msg_context, 'foo', 'foo/Object[]', 'data', serialize, is_numpy)
    compare_file(d, 'object_varlen_deser.txt', result)
    reset_var()
    result = array_serializer_generator(msg_context, 'foo', 'foo/Object[3]', 'data', serialize, is_numpy)
    compare_file(d, 'object_fixed_deser.txt', result)

    # test w/ bad args
    try:
        result = array_serializer_generator(msg_context, '', 'uint8', 'data', True, False)
        exhaust(result)
        assert False, 'should have raised'
    except MsgGenerationException:
        pass


def test_complex_serializer_generator():
    from genmsg.msg_loader import load_msg_by_type
    from genpy.generator import complex_serializer_generator, MsgGenerationException, reset_var
    array_d = os.path.join(get_test_dir(), 'array')
    complex_d = os.path.join(get_test_dir(), 'complex')
    # generator tests are mainly tripwires/coverage tests

    # array_serializer_generator(msg_context, package, type_, name, serialize, is_numpy):
    msg_context = MsgContext.create_default()
    # load in some objects
    search_path = {'foo': [array_d]}
    load_msg_by_type(msg_context, 'foo/Object', search_path)
    load_msg_by_type(msg_context, 'foo/ObjectArray', search_path)

    serialize = True
    is_numpy = False
    reset_var()
    result = complex_serializer_generator(msg_context, 'foo', 'foo/Object', 'data', serialize, is_numpy)
    compare_file(complex_d, 'object_ser.txt', result)
    reset_var()
    result = complex_serializer_generator(msg_context, 'foo', 'foo/Object[]', 'data', serialize, is_numpy)
    compare_file(array_d, 'object_varlen_ser.txt', result)
    reset_var()
    result = complex_serializer_generator(msg_context, 'foo', 'foo/Object[3]', 'data', serialize, is_numpy)
    compare_file(array_d, 'object_fixed_ser.txt', result)

    serialize = False

    reset_var()
    result = complex_serializer_generator(msg_context, 'foo', 'foo/Object[]', 'data', serialize, is_numpy)
    compare_file(array_d, 'object_varlen_deser.txt', result)
    reset_var()
    result = complex_serializer_generator(msg_context, 'foo', 'foo/Object[3]', 'data', serialize, is_numpy)
    compare_file(array_d, 'object_fixed_deser.txt', result)

    try:
        result = complex_serializer_generator(msg_context, 'foo', 'bad/Object', 'data', serialize, is_numpy)
        exhaust(result)
        assert False, 'should have raised'
    except MsgGenerationException:
        pass


def test_serialize_fn_generator():
    from genmsg.msg_loader import load_msg_by_type
    from genpy.generator import serialize_fn_generator, reset_var
    array_d = os.path.join(get_test_dir(), 'array')
    complex_d = os.path.join(get_test_dir(), 'complex')
    # generator tests are mainly tripwires/coverage tests

    # array_serializer_generator(msg_context, package, type_, name, serialize, is_numpy):
    msg_context = MsgContext.create_default()
    # load in some objects
    search_path = {'foo': [array_d]}
    object_spec = load_msg_by_type(msg_context, 'foo/Object', search_path)
    object_array_spec = load_msg_by_type(msg_context, 'foo/ObjectArray', search_path)

    is_numpy = False
    reset_var()
    result = serialize_fn_generator(msg_context, object_spec, is_numpy)
    compare_file(complex_d, 'object_ser_full.txt', result)
    reset_var()
    result = serialize_fn_generator(msg_context, object_array_spec, is_numpy)
    compare_file(array_d, 'object_varlen_ser_full.txt', result)
    reset_var()



from catkin_pkg.python_setup import generate_distutils_setup

from setuptools import setup

d = generate_distutils_setup(
    packages=['genpy'],
    package_dir={'': 'src'},
    requires=['genmsg']
)

setup(**d)



# Software License Agreement (BSD License)
#
# Copyright (c) 2009, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import unittest
import warnings


class RostimeTest(unittest.TestCase):

    def test_TVal(self, TVal=None, test_neg=True):
        import genpy.rostime
        if TVal is None:
            # cannot set as default arg because --cov option reloads the module and causes
            # spurious failure
            TVal = genpy.rostime.TVal
        # test constructor
        # - test zero
        v = TVal()
        self.assert_(repr(v))
        self.assert_(str(v))
        self.assertEqual(0, v.secs)
        self.assertEqual(0, v.nsecs)
        self.assertFalse(v) # test __zero__
        self.assert_(v.is_zero())
        self.assertEqual('0', str(v))
        self.assertEqual(0, v.to_nsec())
        self.assertEqual(0, v.to_sec())

        self.assertEqual(v, v)
        self.assertEqual(v, TVal())
        self.assertEqual(v, TVal(0))
        self.assertEqual(v, TVal(0, 0))
        self.assertEqual(v.__hash__(), TVal(0, 0).__hash__())

        self.assert_(v != TVal(0,1))
        self.assert_(v >= TVal())
        self.assert_(v <= TVal())
        self.assert_(v < TVal(0,1))
        self.assert_(TVal(0,1) > v)
        v.set(0, 0)
        self.assertEqual(0, v.secs)
        self.assertEqual(0, v.nsecs)
        v.set(1, 0)
        self.assertEqual(1, v.secs)
        self.assertEqual(0, v.nsecs)
        v.set(0, 1)
        self.assertEqual(0, v.secs)
        self.assertEqual(1, v.nsecs)
        # - set does _not_ canonicalize
        v.set(0, 1000000000)
        self.assertEqual(0, v.secs)
        self.assertEqual(1000000000, v.nsecs)
        v.canon()
        self.assertEqual(1, v.secs)
        self.assertEqual(0, v.nsecs)

        # - test seconds
        v = TVal(1)
        self.assertEqual(1, v.secs)
        self.assertEqual(0, v.nsecs)
        self.assert_(v) # test __zero__
        self.assertFalse(v.is_zero())
        self.assertEqual('1000000000', str(v))
        self.assertEqual(1000000000, v.to_nsec())
        self.assertEqual(v, v)
        self.assertEqual(v, TVal(1))
        self.assertEqual(v, TVal(1, 0))
        self.assertEqual(v, TVal(0,1000000000))
        self.assertEqual(v.__hash__(), TVal(0,1000000000).__hash__())
        self.assertNotEquals(v, TVal(0, 0))
        self.assertNotEquals(v.__hash__(), TVal(0, 0).__hash__())
        self.assertEqual(NotImplemented, v.__ge__(0))
        class Foo(object): pass
        self.assertEqual(NotImplemented, v.__gt__(Foo()))
        self.assertEqual(NotImplemented, v.__ge__(Foo()))
        self.assertEqual(NotImplemented, v.__le__(Foo()))
        self.assertEqual(NotImplemented, v.__lt__(Foo()))
        self.assertFalse(v.__eq__(Foo()))
        self.assert_(v.__ne__(Foo()))
        self.assert_(v >= TVal())
        self.assert_(v <= TVal(1))
        self.assert_(v <= TVal(1,0))
        self.assert_(v <= TVal(2,0))
        self.assert_(v < TVal(2))
        self.assert_(v < TVal(1,1))
        self.assert_(TVal(1,1) > v)
        self.assert_(TVal(2) > v)
        # - test ns
        v = TVal(0, 1)
        self.assertEqual(0, v.secs)
        self.assertEqual(1, v.nsecs)
        self.assert_(v) # test __zero__
        self.assertFalse(v.is_zero())
        self.assertEqual('1', str(v))
        self.assertEqual(1, v.to_nsec())
        self.assertEqual(v, v)
        self.assertEqual(v, TVal(0,1))
        self.assertNotEquals(v, TVal(0, 0))
        self.assert_(v >= TVal())
        self.assert_(v <= TVal(1))
        self.assert_(v <= TVal(0,1))
        self.assert_(v <= TVal(2,0))
        self.assert_(v < TVal(0,2))
        self.assert_(v < TVal(1))
        self.assert_(TVal(1) > v)
        self.assert_(TVal(0,2) > v)
        # - test canon
        v = TVal(1, 1000000000)
        self.assertEqual(2, v.secs)
        self.assertEqual(0, v.nsecs)
        self.assertEqual(2, v.to_sec())
        self.assertEqual(2000000000, v.to_nsec())

        v = TVal(1, 1000000001)
        self.assertEqual(2, v.secs)
        self.assertEqual(1, v.nsecs)
        self.assertEqual(2.000000001, v.to_sec())
        self.assertEqual(2000000001, v.to_nsec())

        v = TVal(1, -1000000000)
        self.assertEqual(0, v.secs)
        self.assertEqual(0, v.nsecs)
        v = TVal(1, -999999999)
        self.assertEqual(0, v.secs)
        self.assertEqual(1, v.nsecs)
        self.assertEqual(0.000000001, v.to_sec())
        self.assertEqual(1, v.to_nsec())

        if test_neg:
            v = TVal(-1, -1000000000)
            self.assertEqual(-2, v.secs)
            self.assertEqual(0, v.nsecs)
            self.assertEqual(-2, v.to_sec())
            self.assertEqual(-2000000000, v.to_nsec())

            v = TVal(-2, 1000000000)
            self.assertEqual(-1, v.secs)
            self.assertEqual(0, v.nsecs)
            self.assertEqual(-1, v.to_sec())
            self.assertEqual(-1000000000, v.to_nsec())


        # test some more hashes
        self.assertEqual(TVal(1).__hash__(), TVal(1).__hash__())
        self.assertEqual(TVal(1,1).__hash__(), TVal(1,1).__hash__())
        self.assertNotEquals(TVal(1).__hash__(), TVal(2).__hash__())
        self.assertNotEquals(TVal(1,1).__hash__(), TVal(1,2).__hash__())
        self.assertNotEquals(TVal(1,1).__hash__(), TVal(2,1).__hash__())

    def test_Time(self):
        from genpy.rostime import Time, Duration
        self.test_TVal(TVal=Time, test_neg=False)

        # #1600 Duration > Time should fail
        failed = False
        try:
          v = Duration.from_sec(0.1) > Time.from_sec(0.5)
          failed = True
        except: pass
        self.assertFalse(failed, "should have failed to compare")
        try:
          v = Time.from_sec(0.4) > Duration.from_sec(0.1)
          failed = True
        except: pass
        self.assertFalse(failed, "should have failed to compare")

        # TODO: sub

        # neg time fails
        try:
            Time(-1)
            failed = True
        except: pass
        self.assertFalse(failed, "negative time not allowed")
        try:
            Time(1, -1000000001)
            failed = True
        except: pass
        self.assertFalse(failed, "negative time not allowed")

        # test Time.now() is within 10 seconds of actual time (really generous)
        import time
        t = time.time()
        v = Time.from_sec(t)
        self.assertEqual(v.to_sec(), t)
        # test from_sec()
        self.assertEqual(Time.from_sec(0), Time())
        self.assertEqual(Time.from_sec(1.), Time(1))
        self.assertEqual(Time.from_sec(v.to_sec()), v)
        self.assertEqual(v.from_sec(v.to_sec()), v)
        # test to_time()
        self.assertEqual(v.to_sec(), v.to_time())

        # test addition
        # - time + time fails
        try:
            v = Time(1,0) + Time(1, 0)
            failed = True
        except: pass
        self.assertFalse(failed, "Time + Time must fail")

        # - time + duration
        v = Time(1,0) + Duration(1, 0)
        self.assertEqual(Time(2, 0), v)
        v = Duration(1, 0) + Time(1,0)
        self.assertEqual(Time(2, 0), v)
        v = Time(1,1) + Duration(1, 1)
        self.assertEqual(Time(2, 2), v)
        v = Duration(1, 1) + Time(1,1)
        self.assertEqual(Time(2, 2), v)

        v = Time(1) + Duration(0, 1000000000)
        self.assertEqual(Time(2), v)
        v = Duration(1) + Time(0, 1000000000)
        self.assertEqual(Time(2), v)

        v = Time(100, 100) + Duration(300)
        self.assertEqual(Time(400, 100), v)
        v = Duration(300) + Time(100, 100)
        self.assertEqual(Time(400, 100), v)

        v = Time(100, 100) + Duration(300, 300)
        self.assertEqual(Time(400, 400), v)
        v = Duration(300, 300) + Time(100, 100)
        self.assertEqual(Time(400, 400), v)

        v = Time(100, 100) + Duration(300, -101)
        self.assertEqual(Time(399, 999999999), v)
        v =  Duration(300, -101) + Time(100, 100)
        self.assertEqual(Time(399, 999999999), v)

        # test subtraction
        try:
            v = Time(1,0) - 1
            failed = True
        except: pass
        self.assertFalse(failed, "Time - non Duration must fail")
        class Foob(object): pass
        try:
            v = Time(1,0) - Foob()
            failed = True
        except: pass
        self.assertFalse(failed, "Time - non TVal must fail")

        # - Time - Duration
        v = Time(1,0) - Duration(1, 0)
        self.assertEqual(Time(), v)

        v = Time(1,1) - Duration(-1, -1)
        self.assertEqual(Time(2, 2), v)
        v = Time(1) - Duration(0, 1000000000)
        self.assertEqual(Time(), v)
        v = Time(2) - Duration(0, 1000000000)
        self.assertEqual(Time(1), v)
        v = Time(400, 100) - Duration(300)
        self.assertEqual(Time(100, 100), v)
        v = Time(100, 100) - Duration(0, 101)
        self.assertEqual(Time(99, 999999999), v)

        # - Time - Time = Duration
        v = Time(100, 100) - Time(100, 100)
        self.assertEqual(Duration(), v)
        v = Time(100, 100) - Time(100)
        self.assertEqual(Duration(0,100), v)
        v = Time(100) - Time(200)
        self.assertEqual(Duration(-100), v)

        # Time (float secs) vs. Time(int, int)
        self.assertEqual(Time.from_sec(0.5), Time(0.5))
        t = Time(0.5)
        self.assert_(type(t.secs) == int)
        self.assertEqual(0, t.secs)
        self.assertEqual(500000000, t.nsecs)

        try:
            Time(0.5, 0.5)
            self.fail("should have thrown value error")
        except ValueError: pass

    def test_Duration(self):
        from genpy.rostime import Time, Duration

        self.test_TVal(TVal=Duration, test_neg=True)

        # test from_sec
        v = Duration(1000)
        self.assertEqual(v, Duration.from_sec(v.to_sec()))
        self.assertEqual(v, v.from_sec(v.to_sec()))
        v = Duration(0,1000)
        self.assertEqual(v, Duration.from_sec(v.to_sec()))
        self.assertEqual(v, v.from_sec(v.to_sec()))

        # test neg
        v = -Duration(1, -1)
        self.assertEqual(-1, v.secs)
        self.assertEqual(1, v.nsecs)
        v = -Duration(-1, -1)
        self.assertEqual(1, v.secs)
        self.assertEqual(1, v.nsecs)
        v = -Duration(-1, 1)
        self.assertEqual(0, v.secs)
        self.assertEqual(999999999, v.nsecs)

        # test addition
        self.assertEqual(Duration(1,0) + Time(1, 0), Time(2, 0))
        failed = False
        try:
            v = Duration(1,0) + 1
            failed = True
        except: pass
        self.assertFalse(failed, "Duration + int must fail")

        v = Duration(1,0) + Duration(1, 0)
        self.assertEqual(2, v.secs)
        self.assertEqual(0, v.nsecs)
        self.assertEqual(Duration(2, 0), v)
        v = Duration(-1,-1) + Duration(1, 1)
        self.assertEqual(0, v.secs)
        self.assertEqual(0, v.nsecs)
        self.assertEqual(Duration(), v)
        v = Duration(1) + Duration(0, 1000000000)
        self.assertEqual(2, v.secs)
        self.assertEqual(0, v.nsecs)
        self.assertEqual(Duration(2), v)
        v = Duration(100, 100) + Duration(300)
        self.assertEqual(Duration(400, 100), v)
        v = Duration(100, 100) + Duration(300, 300)
        self.assertEqual(Duration(400, 400), v)
        v = Duration(100, 100) + Duration(300, -101)
        self.assertEqual(Duration(399, 999999999), v)

        # test subtraction
        try:
            v = Duration(1,0) - 1
            failed = True
        except: pass
        self.assertFalse(failed, "Duration - non duration must fail")
        try:
            v = Duration(1, 0) - Time(1,0)
            failed = True
        except: pass
        self.assertFalse(failed, "Duration - Time must fail")

        v = Duration(1,0) - Duration(1, 0)
        self.assertEqual(Duration(), v)
        v = Duration(-1,-1) - Duration(1, 1)
        self.assertEqual(Duration(-3, 999999998), v)
        v = Duration(1) - Duration(0, 1000000000)
        self.assertEqual(Duration(), v)
        v = Duration(2) - Duration(0, 1000000000)
        self.assertEqual(Duration(1), v)
        v = Duration(100, 100) - Duration(300)
        self.assertEqual(Duration(-200, 100), v)
        v = Duration(100, 100) - Duration(300, 101)
        self.assertEqual(Duration(-201, 999999999), v)

        # test abs
        self.assertEqual(abs(Duration()), Duration())
        self.assertEqual(abs(Duration(1)), Duration(1))
        self.assertEqual(abs(Duration(-1)), Duration(1))
        self.assertEqual(abs(Duration(0,-1)), Duration(0,1))
        self.assertEqual(abs(Duration(-1,-1)), Duration(1,1))
        self.assertEqual(abs(Duration(0,1)), Duration(0,1))

        # Duration (float secs) vs. Duration(int, int)
        self.assertEqual(Duration.from_sec(0.5), Duration(0.5))
        t = Duration(0.5)
        self.assert_(type(t.secs) == int)
        self.assertEqual(0, t.secs)
        self.assertEqual(500000000, t.nsecs)

        try:
            Duration(0.5, 0.5)
            self.fail("should have thrown value error")
        except ValueError: pass

        # Test mul
        self.assertEqual(Duration(4), Duration(2) * 2)
        self.assertEqual(Duration(4), Duration(2) * 2.)
        self.assertEqual(Duration(4), 2 * Duration(2))
        self.assertEqual(Duration(4), 2. * Duration(2))
        self.assertEqual(Duration(10), Duration(4) * 2.5)
        self.assertEqual(Duration(4, 8), Duration(2, 4) * 2)
        v = Duration(4, 8) - (Duration(2, 4) * 2.)
        self.assert_(abs(v.to_nsec()) < 100)
        v = Duration(5, 10) - (Duration(2, 4) * 2.5)
        self.assert_(abs(v.to_nsec()) < 100)

        # Test div
        self.assertEqual(Duration(4), Duration(8) / 2)
        self.assertEqual(Duration(4), Duration(8) / 2.)
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            self.assertEqual(Duration(4), Duration(8) // 2)
            self.assertEqual(Duration(4), Duration(8) // 2.)
            self.assertEqual(Duration(4), Duration(9) // 2)
            self.assertEqual(Duration(4), Duration(9) // 2.)
            self.assertEqual(len(w), 0)
        self.assertEqual(Duration(4, 2), Duration(8, 4) / 2)
        v = Duration(4, 2) - (Duration(8, 4) / 2.)
        self.assert_(abs(v.to_nsec()) < 100)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            self.assertEqual(Duration(4, 0), Duration(8, 4) // 2)
            self.assertEqual(Duration(4, 0), Duration(9, 5) // 2)
            v = Duration(4, 2) - (Duration(9, 5) // 2.)
            self.assertTrue(abs(v.to_nsec()) < 100)
            self.assertEqual(len(w), 0)
